package core

import (
	"errors"
	"fmt"
	"math/big"
	"math/rand"
	"sync"
	"time"
	"validator/conf"
	"validator/quest"
	"validator/tree"
	"validator/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/ethdb/leveldb"
	"github.com/ethereum/go-ethereum/trie"
	validatorv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/validator/v1"
	log "github.com/sirupsen/logrus"
)

type Validator struct {
	lvdb               *leveldb.Database
	db                 *trie.Database
	state              *StateDB
	pendingProof       map[int64]map[common.Address]*validatorv1.ValidatedProof
	pendingWorkload    uint64
	mtTreeCache        map[string]*tree.MerkleTree
	mstTreeCache       map[string]*tree.MerkleSumTree
	date               string
	rpc                *ChainRPC
	containerAddresses []common.Address
	nmAddresses        []common.Address
	q                  *quest.Quest

	cfg *conf.Config

	sync.Mutex
}

func RunValidator(q *quest.Quest, cfg *conf.Config) *Validator {
	_rpc := newChain(cfg.ChainRPC, cfg.PrivateKey, cfg.StoreContract, cfg.ValidatorContract)
	diskDB, err := leveldb.New(fmt.Sprintf("%s/db", cfg.DataDir), 128, 1024, "", false)
	if err != nil {
		panic(err)
	}

	lastDay, stateRoot := getDBLastDayRoot(diskDB)
	stateDB, err := NewStateDB(diskDB, stateRoot)
	if err != nil {
		panic(err)
	}
	v := &Validator{
		lvdb:         diskDB,
		state:        stateDB,
		pendingProof: make(map[int64]map[common.Address]*validatorv1.ValidatedProof),
		mtTreeCache:  make(map[string]*tree.MerkleTree),
		mstTreeCache: make(map[string]*tree.MerkleSumTree),
		rpc:          _rpc,
		date:         lastDay,
		cfg:          cfg,
		q:            q,
	}

	go v.UpdateContractAddressJob()
	go v.UpdateGlobalWorkloadJob()
	<-time.After(time.Second * 3)
	if v.date != "" {
		// 当前5号，假设db 1号
		// v.date = 1或2时，需要处理的是 2或3号
		for v.dateToTimestamp(v.date) < v.yesterdayTimestamp()-86400 {
			if err = v.SyncDayJob(v.dateToTimestamp(v.date)+86400, true); err != nil {
				return nil
			}
		}
		// v.date = 3， 如果当前时间超过提交昨天的时间，需要提交昨天的
		if v.dateToTimestamp(v.date) < v.yesterdayTimestamp() &&
			time.Now().UTC().After(time.Unix(v.todayTimestamp()+int64(cfg.CommitOffset), 0)) {
			if err = v.SyncDayJob(v.dateToTimestamp(v.date)+86400, true); err != nil {
				return nil
			}
		}
		lastDay, _ = getDBLastDayRoot(diskDB)
	}

	v.LoadMerkleTree(lastDay)
	v.LoadMerkleSumTree(lastDay)
	go v.Ticker()
	return v
}

func getDBLastDayRoot(diskDB *leveldb.Database) (lastDay string, stateRoot common.Hash) {
	ok, err := diskDB.Has([]byte("lastday"))
	if err != nil {
		panic(err)
	}
	if ok {
		val, _ := diskDB.Get([]byte("lastday"))
		log.WithField("last day", string(val)).Info("diskdb get last day")
		lastDay = string(val)
		root, _ := diskDB.Get([]byte(fmt.Sprintf("sroot:%s", string(val))))
		stateRoot = common.BytesToHash(root)
		log.WithFields(log.Fields{
			"last day":   lastDay,
			"state root": stateRoot.Hex(),
		}).Info("load last day state from diskdb")
	}
	return lastDay, stateRoot
}

func (v *Validator) AddPendingProof(miner common.Address, proof *validatorv1.ValidatedProof) {
	v.Lock()
	defer v.Unlock()

	var userTodayWorkload uint64
	currentPendingProof, ok := v.pendingProof[v.todayTimestamp()][miner]
	if ok {
		userTodayWorkload = currentPendingProof.Workload
	}

	if v.pendingProof[v.todayTimestamp()] == nil {
		v.pendingProof[v.todayTimestamp()] = make(map[common.Address]*validatorv1.ValidatedProof)
	}
	v.pendingProof[v.todayTimestamp()][miner] = &validatorv1.ValidatedProof{Workload: userTodayWorkload + proof.Workload}

	log.WithFields(log.Fields{
		"miner":            miner.Hex(),
		"current_workload": userTodayWorkload + proof.Workload,
		"workload":         proof.Workload,
		"global_workload":  v.pendingWorkload,
	}).Debug("add validated proof")
}

// ProcessDayJob 每日任务
func (v *Validator) ProcessDayJob() {
	dbLastDay, _ := getDBLastDayRoot(v.lvdb)
	if dbLastDay == v.yesterdayString() && v.date != "" {
		log.WithField("date", v.date).Warn("already process day job")
		return
	}
	log.WithFields(log.Fields{
		"today":               v.todayString(),
		"today timestamp":     v.todayTimestamp(),
		"yesterday":           v.yesterdayString(),
		"yesterday timestamp": v.yesterdayTimestamp(),
	}).Debug("process day job")
	v.LoadPendingProofs(v.yesterdayTimestamp(), v.todayTimestamp())
	v.date = v.yesterdayString()
	dayProof, totalWorkload := v.Commit()
	mstRoot, _, err := v.CommitMST(dayProof)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle sum tree")
		return
	}
	objects := v.state.IterAllObject()
	mtRoot, err := v.CommitMT(objects)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle tree")
		return
	}
	txHash, err := v.rpc.SubmitProofs(v.dateToTimestamp(v.date), mstRoot, mtRoot, totalWorkload)
	if err != nil {
		log.WithError(err).Error("submit proofs")
		return
	}
	err = v.lvdb.Put([]byte(fmt.Sprintf("txid:%s", v.date)), txHash.Bytes())
	if err != nil {
		log.WithError(err).Error("db failed to save txid")
		return
	}
	log.WithField("tx hash", txHash.Hex()).Info("submit proofs to contract")
	err = v.lvdb.Put([]byte("lastday"), []byte(v.date))
	if err != nil {
		log.WithError(err).Error("db failed to update last day")
		return
	}
	log.Info("process day done")
}

// SyncDayJob 同步任务
func (v *Validator) SyncDayJob(dateTimestamp int64, commitToChain bool) (err error) {
	dbLastDay, _ := getDBLastDayRoot(v.lvdb)
	if dbLastDay == v.yesterdayString() && v.date != "" {
		log.WithField("date", v.date).Warn("already process day job")
		return errors.New("already process day job")
	}
	log.WithFields(log.Fields{
		"date":           v.timestampToDate(dateTimestamp),
		"date timestamp": dateTimestamp,
	}).Debug("sync day job")
	v.LoadPendingProofs(dateTimestamp, dateTimestamp+86400)
	v.date = v.timestampToDate(dateTimestamp)
	dayProof, totalWorkload := v.Commit()
	mstRoot, _, err := v.CommitMST(dayProof)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle sum tree")
		return
	}
	objects := v.state.IterAllObject()
	mtRoot, err := v.CommitMT(objects)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle tree")
		return
	}
	var txHash common.Hash
	if commitToChain {
		txHash, err = v.rpc.SubmitProofs(dateTimestamp, mstRoot, mtRoot, totalWorkload)
		if err != nil {
			log.WithError(err).Error("submit proofs")
			return
		}
	}
	err = v.lvdb.Put([]byte(fmt.Sprintf("txid:%s", v.date)), txHash.Bytes())
	if err != nil {
		log.WithError(err).Error("db failed to save txid")
		return
	}
	log.WithField("tx hash", txHash.Hex()).Info("submit proofs to contract")
	err = v.lvdb.Put([]byte("lastday"), []byte(v.date))
	if err != nil {
		log.WithError(err).Error("db failed to update last day")
		return
	}
	log.Info("sync day done")
	return
}

// Commit statedb提交
func (v *Validator) Commit() (dayProofs map[common.Address]*validatorv1.ValidatedProof, totalWorkload uint64) {
	st := time.Now()
	proof, totalWorkload := v.RefreshPendingProof()
	balancePerWorkload, err := v.rpc.GetWorkloadThreshold(totalWorkload)
	if err != nil {
		log.WithError(err).Error("failed to get workload threshold")
		return
	}
	log.WithFields(log.Fields{
		"workload":                 totalWorkload,
		"amount per workload(wei)": balancePerWorkload.String(),
	}).Info("get workload threshold from contract")

	for miner, proof := range proof {
		balance := big.NewInt(0).Mul(balancePerWorkload, big.NewInt(int64(proof.Workload)))
		proof.Balance = balance.String()
		err := v.SealProof(miner, proof)
		if err != nil {
			log.WithError(err).Error("failed to seal proof")
			return
		}
	}
	root, err := v.state.Commit()
	if err != nil {
		return
	}
	log.WithFields(log.Fields{"root": root.Hex(), "cost": time.Since(st).String()}).Info("commit state root")
	err = v.lvdb.Put([]byte(fmt.Sprintf("sroot:%s", v.date)), root.Bytes())
	if err != nil {
		log.WithError(err).Error("failed to update state root")
		return
	}
	v.state, _ = NewStateDB(v.lvdb, root)
	return proof, totalWorkload
}

func (v *Validator) SealProof(miner common.Address, proof *validatorv1.ValidatedProof) (err error) {
	minerObject := v.state.GetMinerObject(miner)
	oldBalance := minerObject.Balance
	minerObject.Balance = util.MustAddStringToBigInt(oldBalance, proof.Balance).String()
	err = v.state.UpdateMinerObject(miner, minerObject)
	if err != nil {
		return err
	}
	log.WithFields(log.Fields{
		"miner":       miner.Hex(),
		"old_balance": oldBalance,
		"new_balance": minerObject.Balance,
	}).Info("seal proof")
	return err
}

func (v *Validator) RefreshPendingProof() (proof map[common.Address]*validatorv1.ValidatedProof, totalWorkload uint64) {
	v.Lock()
	defer v.Unlock()
	proof = make(map[common.Address]*validatorv1.ValidatedProof)
	for timestamp, pendingProofMap := range v.pendingProof {
		if timestamp != v.todayTimestamp() {
			continue
		}
		for miner, pendingProof := range pendingProofMap {
			proof[miner] = pendingProof
			totalWorkload += pendingProof.Workload
		}
	}
	v.pendingProof[v.yesterdayTimestamp()] = make(map[common.Address]*validatorv1.ValidatedProof)
	return
}

func (v *Validator) Ticker() {
	executionTime := time.Unix(v.todayTimestamp()+86400+int64(v.cfg.CommitOffset), 0)
	// executionTime := time.Now().Add(time.Second * 3) for test
	waitTime := executionTime.Sub(time.Now().UTC())
	timer := time.NewTimer(waitTime)
	log.WithFields(log.Fields{
		"execution_time": executionTime.UTC().String(),
		"wait_time":      waitTime.String(),
	}).Info("prepare commit task")
	for {
		<-timer.C
		v.ProcessDayJob()
		executionTime = executionTime.Add(v.duration())
		waitTime = executionTime.Sub(time.Now().UTC())
		timer.Reset(waitTime)
		log.WithFields(log.Fields{
			"execution_time": executionTime.UTC().String(),
			"wait_time":      waitTime.String(),
		}).Info("prepare commit task")
	}
}

func (v *Validator) Mock() {
	addresses := make([]common.Address, 10)
	rand.Seed(time.Now().Unix())
	for i := 0; i < 10; i++ {
		b20 := make([]byte, 20)
		rand.Read(b20)
		addresses[i] = common.BytesToAddress(b20)
	}

	for {
		time.Sleep(time.Second)
		v.AddPendingProof(addresses[rand.Intn(10)], &validatorv1.ValidatedProof{
			Workload:  uint64((rand.Int63n(9) + 1) * 100),
			Timestamp: uint64(time.Now().Unix() - 1),
		})
	}
}

func (v *Validator) IsContainerAddress(address common.Address) bool {
	for _, addr := range v.containerAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (v *Validator) IsNMAddress(address common.Address) bool {
	for _, addr := range v.nmAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (v *Validator) todayString() string {
	return time.Unix(v.todayTimestamp(), 0).Format("2006-01-02")
}

func (v *Validator) yesterdayString() string {
	return time.Unix(v.yesterdayTimestamp(), 0).Format("2006-01-02")
}

func (v *Validator) todayTimestamp() int64 {
	now := time.Now().UTC()
	today := time.Date(now.Year(), now.Month(), now.Day(), 0, 0, 0, 0, time.UTC)
	// today := time.Date(now.Year(), now.Month(), now.Day()+1, 0, 0, 0, 0, time.UTC) for test
	return today.Unix()
}

func (v *Validator) yesterdayTimestamp() int64 {
	return v.todayTimestamp() - 86400
}

func (v *Validator) dateToTimestamp(date string) int64 {
	t, err := time.ParseInLocation("2006-01-02", date, time.UTC)
	if err != nil {
		return 0
	}
	return t.Unix()
}

func (v *Validator) timestampToDate(timestamp int64) string {
	return time.Unix(timestamp, 0).Format("2006-01-02")
}

func (v *Validator) duration() time.Duration {
	return 3600 * 24 * time.Second
}
